/// @file
////////////////////////////////////////////////////////////////////////////////////////////////////
///
/// Copyright (C) 2019/20      Christian Lessig, Otto-von-Guericke Universitaet Magdeburg
///
////////////////////////////////////////////////////////////////////////////////////////////////////
///
///  module     : lecture 10
///
///  author     : lessig@isg.cs.uni-magdeburg.de
///
///  project    : GPU Programming
///
///  description: simple Cuda demo
///
////////////////////////////////////////////////////////////////////////////////////////////////////

// includes, system
#include <iostream>

// includes, project
#include "cuda_util.h"

////////////////////////////////////////////////////////////////////////////////////////////////////
// add two vectors together
////////////////////////////////////////////////////////////////////////////////////////////////////
__global__ void
vecAdd( const float* a, const float* b, float* c, unsigned int n) {

  const unsigned int tid = blockIdx.x * blockDim.x + threadIdx.x;

  if( tid < n) {
    c[tid] = a[tid] + b[tid];
  }
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize Cuda device
////////////////////////////////////////////////////////////////////////////////////////////////////
void
initDevice( int& device_handle) {

  int deviceCount = 0;
  cudaGetDeviceCount(&deviceCount);

  if( 0 == deviceCount) {
    std::cerr << "initDevice() : No CUDA device found." << std::endl;
  }

  // one could implement more complex logic here to find the fastest device
  if( deviceCount > 1) {
    std::cerr << "initDevice() : Multiple CUDA devices found. Using first one." << std::endl;
  }

  // set the device
  cudaSetDevice( device_handle);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
initDeviceMemory( const float* a, const float* b, const unsigned int n,
                  float*& a_device, float*& b_device, float*& c_device)
{
  // allocate device memory
  cudaMalloc((void **) &a_device, sizeof(float) * n);
  cudaMalloc((void **) &b_device, sizeof(float) * n);
  cudaMalloc((void **) &c_device, sizeof(float) * n);

  // copy device memory
  cudaMemcpy( a_device, a, sizeof(float) * n, cudaMemcpyHostToDevice );
  cudaMemcpy( b_device, b, sizeof(float) * n, cudaMemcpyHostToDevice );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// initialize device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
getResultDevice( const float* c_device, float* c, const unsigned int n) {

  cudaMemcpy( c, c_device, sizeof(float) * n, cudaMemcpyDeviceToHost );
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// free device memory
////////////////////////////////////////////////////////////////////////////////////////////////////
void
freeDeviceMemory( float*& a_device, float*& b_device, float*& c_device) {

  cudaFree( a_device);
  cudaFree( b_device);
  cudaFree( c_device);
}

////////////////////////////////////////////////////////////////////////////////////////////////////
// program entry point
////////////////////////////////////////////////////////////////////////////////////////////////////
int
main( int /*argc*/, char** /*argv*/ ) {

  // initialize device
  int device_handle = 0;
  initDevice( device_handle);

  // set up host memory
  const unsigned int n = 2 << 15;
  float* a = (float*) malloc( sizeof(float) * n);
  float* b = (float*) malloc( sizeof(float) * n);
  float* c = (float*) malloc( sizeof(float) * n);
  for( unsigned int i = 0; i < n; ++i) {
    a[i] = (float) i;
    b[i] = (float) i;
  }

  // initialize device memory
  float* a_device = nullptr;
  float* b_device = nullptr;
  float* c_device = nullptr;
  initDeviceMemory( a, b, n, a_device, b_device, c_device);

  // clean host memory that is no longer needed
  free( a);
  free( b);

  // determine thread layout
  const unsigned int MAX_THREADS_PER_BLOCK = 1024;
  int num_threads_per_block = MAX_THREADS_PER_BLOCK;
  int num_blocks = n / MAX_THREADS_PER_BLOCK;
  if( 0 != n % MAX_THREADS_PER_BLOCK) {
    num_blocks++;
  }

  // run kernel
  vecAdd<<< num_blocks , num_threads_per_block >>>( a_device, b_device, c_device, n);
  cudaDeviceSynchronize();

  // copy result back to host
  getResultDevice( c_device, c, n);

  // clean up device memory
  freeDeviceMemory( a_device, b_device, c_device);

  // clean up host memory
  free( c);

  return EXIT_SUCCESS;
}
